##############################################################################
#
# Copyright (c) 2004 Zope Corporation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
from __future__ import with_statement

import doctest
import logging
import manuel.capture
import manuel.doctest
import manuel.testing
import sys
import threading
import time
import unittest
import warnings
import zc.ngi.async
import zc.ngi.generator
import zc.ngi.testing
import zc.ngi.wordcount

def blocking_warns():
    """
    >>> assert_(len(blocking_warnings) == 1)
    >>> assert_(blocking_warnings[-1].category is DeprecationWarning)
    >>> print blocking_warnings[-1].message
    The blocking module is deprecated.
    """

if sys.version_info >= (2, 6):
    # silence blocking deprecation warning
    with warnings.catch_warnings(record=True) as blocking_warnings:
        warnings.simplefilter('default')
        # omg, without record=True, warnings aren't actually caught.
        # Who thinks up this stuff?
        import zc.ngi.blocking
else:
    del blocking_warns



def test_async_cannot_connect():
    """Let's make sure that the connector handles connection failures correctly

    >>> import threading
    >>> lock = threading.Lock()
    >>> _ = lock.acquire()

    We define a simple handler that just notifies of failed connections.

    >>> class Handler:
    ...     def failed_connect(connection, reason):
    ...         print 'failed'
    ...         lock.release()

    >>> def connect(addr):
    ...     zc.ngi.async.connect(addr, Handler())
    ...     lock.acquire()

    We find an unused port (so when we connect to it, the connection
    will fail).

    >>> port = zc.ngi.testing.get_port()

    Now let's try to connect

    >>> connect(('localhost', port))
    failed

    """

def async_thread_management():
    """

    There's no thread by default:

    >>> len([t for t in threading.enumerate() if t.getName() == 'zc.ngi.async'])
    0

    There's a default name:

    >>> listener = zc.ngi.async.listener(None, lambda _: None)
    >>> len([t for t in threading.enumerate() if t.getName() == 'zc.ngi.async'])
    1
    >>> listener.close()
    >>> zc.ngi.async.wait(1)

    When there's nothing to do, the thread goes away:

    >>> len([t for t in threading.enumerate() if t.getName() == 'zc.ngi.async'])
    0

    If we create out own implementation, we can give it a name:

    >>> impl = zc.ngi.async.Implementation(name='bob')
    >>> listener = impl.listener(None, lambda _: None)
    >>> len([t for t in threading.enumerate() if t.getName() == 'bob'])
    1
    >>> listener.close()
    >>> impl.wait(1)

    Otherwise, it gets a slightly more descriptive name:

    >>> impl = zc.ngi.async.Implementation('')
    >>> listener = impl.listener(None, lambda _: None)
    >>> len([t for t in threading.enumerate()
    ...     if t.getName() == 'zc.ngi.async application created'])
    1
    >>> listener.close()
    >>> impl.wait(1)

    """

def blocking_connector_handles_failed_connect():
    """
    >>> import zc.ngi.blocking
    >>> zc.ngi.blocking.open(('localhost', 42), zc.ngi.testing.connect)
    Traceback (most recent call last):
    ...
    ConnectionFailed: no such server

    """

def failure_to_bind_removes_listener_from_socket_map():
    """
    First, grab a port:

    >>> import socket, random

    >>> s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    >>> for i in range(1000):
    ...    port = random.randint(10000, 30000)
    ...    try: s.bind(('', port))
    ...    except socket.error: pass
    ...    else: break
    ... else: print 'woops'

    Make sure thread is running by creating a listener

    >>> place_holder = zc.ngi.async.listener(None, lambda _: None)
    >>> time.sleep(.1)

    Get size of socket map:

    >>> len(zc.ngi.async._map)
    2

    Now, trying to create a listener on the port should fail, and the
    map should remain the same size.

    >>> try: zc.ngi.async.listener(('', port), None)
    ... except socket.error: pass
    ... else: print 'oops'

    >>> len(zc.ngi.async._map)
    2

    >>> s.close()

    UDP:

    >>> s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    >>> for i in range(1000):
    ...    port = random.randint(10000, 30000)
    ...    try: s.bind(('', port))
    ...    except socket.error: pass
    ...    else: break
    ... else: print 'woops'

    Get size of socket map:

    >>> len(zc.ngi.async._map)
    2

    Now, trying to create a listener on the port should fail, and the
    map should remain the same size.

    >>> try: zc.ngi.async.udp_listener(('', port), None)
    ... except socket.error: pass
    ... else: print 'oops'

    >>> len(zc.ngi.async._map)
    2

    >>> s.close()
    >>> place_holder.close()
    """

def async_error_in_client_when_conection_is_closed():
    """
If a connection is closed, we need to make sure write calls generate errors.

    >>> logger = logging.getLogger('zc.ngi')
    >>> log_handler = logging.StreamHandler(sys.stdout)
    >>> logger.addHandler(log_handler)
    >>> logger.setLevel(logging.WARNING)

    >>> server_event = threading.Event()
    >>> @zc.ngi.generator.handler
    ... def server(conn):
    ...     data = yield
    ...     print data
    ...     server_event.set()

    >>> listener = zc.ngi.async.listener(None, server)

    >>> class Connector:
    ...     def __init__(self):
    ...         self.event = threading.Event()
    ...     def connected(self, conn):
    ...         self.conn = conn
    ...         self.event.set()

    >>> connector = Connector()
    >>> zc.ngi.async.connect(listener.address, connector)
    >>> _ = connector.event.wait(1)

OK, we've connected.  If we close the connection, we won't be able to write:

    >>> connector.conn.close()
    >>> connector.conn.write('xxx')

    >>> connector.conn.writelines(['xxx', 'yyy'])

Similarly if the server closes the connection:

    >>> connector = Connector()
    >>> zc.ngi.async.connect(listener.address, connector)
    >>> _ = connector.event.wait(1)

    >>> connector.conn.write('aaa'); _ = server_event.wait(1)
    aaa

    >>> connector.conn.write('xxx')

    >>> connector.conn.writelines(['xxx', 'yyy'])


    >>> logger.removeHandler(log_handler)
    >>> logger.setLevel(logging.NOTSET)

    >>> listener.close()
    """

def when_a_server_closes_a_connection_blocking_request_returns_reason():
    """

    >>> import zc.ngi.adapters, zc.ngi.async, zc.ngi.blocking
    >>> @zc.ngi.adapters.Sized.handler
    ... def echo1(c):
    ...     c.write((yield))

    >>> listener = zc.ngi.async.listener(None, echo1)
    >>> @zc.ngi.adapters.Sized.handler
    ... def client(c):
    ...     c.write('test')
    ...     print '1', (yield)
    ...     print '2', (yield)
    >>> zc.ngi.blocking.request(zc.ngi.async.connect, listener.address,
    ...                         client, 1)
    ... # doctest: +ELLIPSIS
    1...
    'end of input'
    >>> listener.close()
    """

def errors_raised_by_handler_should_be_propigated_by_blocking_request():
    """
    Errors raised by handlers should propigate to the request caller,
    rather than just getting logged as usual.

    Note that this test also exercises error handling in zc.ngi.async.

    >>> from zc.ngi import async
    >>> from zc.ngi.adapters import Sized
    >>> from zc.ngi.blocking import request

    >>> @Sized.handler
    ... def echo(c):
    ...     while 1:
    ...         data = (yield)
    ...         if data == 'stop': break
    ...         c.write(data)

    >>> listener = async.listener(None, echo)

    Handle error in setup

    >>> @Sized.handler
    ... def bad(c):
    ...     raise ValueError

    >>> try: request(async.connect, listener.address, bad, 1)
    ... except ValueError: pass
    ... else: print 'oops'

    Handle error in input

    >>> @Sized.handler
    ... def bad(c):
    ...     c.write('test')
    ...     data = (yield)
    ...     raise ValueError

    >>> try: request(async.connect, listener.address, bad, 1)
    ... except ValueError: pass
    ... else: print 'oops'

    Handle error in close

    >>> @Sized.handler
    ... def bad(c):
    ...     c.write('stop')
    ...     try:
    ...         while 1:
    ...             data = (yield)
    ...     except GeneratorExit:
    ...         raise ValueError

    >>> try: request(async.connect, listener.address, bad, 1)
    ... except ValueError: pass
    ... else: print 'oops'

    Handle error in handle_exception arising from error during iteration:

    >>> @Sized.handler
    ... def bad(c):
    ...     c.writelines(XXX for i in range(2))
    ...     data = (yield)

    >>> try: request(async.connect, listener.address, bad, 1)
    ... except NameError: pass
    ... else: print 'oops'

    >>> listener.close()
    """

def async_handling_iteration_errors():
    """

    >>> from zc.ngi import async
    >>> from zc.ngi.adapters import Sized
    >>> from zc.ngi.blocking import request

    >>> @Sized.handler
    ... def echo(c):
    ...     while 1:
    ...         data = (yield)
    ...         if data == 'stop': break
    ...         c.write(data)

    >>> listener = async.listener(None, echo)

    Handler with no handle_exception but with a handle close.

    >>> event = threading.Event()
    >>> class Bad:
    ...    def connected(self, connection):
    ...        connection.set_handler(self)
    ...        connection.writelines(XXX for i in range(2))
    ...    def handle_close(self, connection, reason):
    ...        print 'closed', reason
    ...        event.set()

    >>> zc.ngi.async.connect(listener.address, Bad()); _ = event.wait(1)
    closed Bad instance has no attribute 'handle_exception'

    >>> listener.close()
    """

def assert_(cond, *args):
    if not cond:
        raise AssertionError(*args)

def setHandler_compatibility():
    """
Make sure setHandler still works, with deprecation warnings:

The testing connection warns:

    >>> class Handler:
    ...     def handle_input(self, connection, data):
    ...         print 'got', `data`

    >>> conn = zc.ngi.testing.Connection()
    >>> with warnings.catch_warnings(record=True) as caught:
    ...     warnings.simplefilter('default')
    ...     conn.setHandler(Handler())
    ...     assert_(len(caught) == 1, len(caught))
    ...     assert_(caught[-1].category is DeprecationWarning)
    ...     print caught[-1].message
    setHandler is deprecated. Use set_handler,

    >>> conn.test_input('test')
    got 'test'

The async connections warn:

    >>> server_event = threading.Event()
    >>> class server:
    ...     def __init__(self, c):
    ...         global server_caught
    ...         with warnings.catch_warnings(record=True) as caught:
    ...             warnings.simplefilter('default')
    ...             c.setHandler(self)
    ...             server_caught = caught
    ...         server_event.set()
    ...         c.close()

    >>> client_event = threading.Event()
    >>> class client:
    ...
    ...     def connected(self, c):
    ...         global client_caught
    ...         with warnings.catch_warnings(record=True) as caught:
    ...             warnings.simplefilter('default')
    ...             c.setHandler(self)
    ...             client_caught = caught
    ...         client_event.set()
    ...         c.close()

    >>> listener = zc.ngi.async.listener(None, server)
    >>> zc.ngi.async.connect(listener.address, client())
    >>> _ = server_event.wait(1)
    >>> _ = client_event.wait(1)
    >>> listener.close()

    >>> assert_(len(server_caught) == 1)
    >>> assert_(server_caught[0].category is DeprecationWarning)
    >>> print server_caught[0].message
    setHandler is deprecated. Use set_handler,

    >>> assert_(len(client_caught) == 1)
    >>> assert_(client_caught[0].category is DeprecationWarning)
    >>> print client_caught[0].message
    setHandler is deprecated. Use set_handler,

    >>> zc.ngi.async.wait(1)

The adapters warn:

    >>> import zc.ngi.adapters
    >>> with warnings.catch_warnings(record=True) as caught:
    ...     warnings.simplefilter('default')
    ...     conn = zc.ngi.adapters.Lines(zc.ngi.testing.Connection())
    ...     conn.setHandler(Handler())
    ...     assert_(len(caught) == 1)
    ...     assert_(caught[-1].category is DeprecationWarning)
    ...     print caught[-1].message
    setHandler is deprecated. Use set_handler,

    >>> class OldConn:
    ...     def setHandler(self, h):
    ...         print 'setHandler called'
    ...         global old_handler
    ...         old_handler = h

    >>> with warnings.catch_warnings(record=True) as caught:
    ...     warnings.simplefilter('default')
    ...     conn = zc.ngi.adapters.Lines(OldConn())
    ...     handler = Handler()
    ...     conn.set_handler(handler)
    ...     assert_(len(caught) == 1)
    ...     assert_(caught[-1].category is DeprecationWarning)
    ...     print caught[-1].message
    ...     assert_(old_handler is conn)
    setHandler called
    setHandler is deprecated. Use set_handler,


    """

if sys.version_info < (2, 6):
    del setHandler_compatibility

class BrokenConnect:

    connected = failed_connect = __call__ = lambda: xxxxx

class BrokenAfterConnect:

    def connected(self, connection):
        connection.write("Hee hee\0")
        connection.set_handler(self)

    __call__ = connected

    handle_input = handle_close = lambda: xxxxx

def async_evil_setup(test):

    # Uncomment the next 2 lines to check that a bunch of lambda type
    # errors are logged.
    #import logging
    #logging.getLogger().addHandler(logging.StreamHandler())

    # clean up the map.
    zc.ngi.async.cleanup_map()

    # See if we can break the main loop before running the async test

    # Connect to bad port with bad handler

    port = zc.ngi.wordcount.get_port()
    addr = 'localhost', port
    zc.ngi.async.connect(addr, BrokenConnect())

    # Start the server and connect to a good port with a bad handler

    port = zc.ngi.wordcount.start_server_process(zc.ngi.async.listener)
    addr = 'localhost', port
    zc.ngi.async.connect(addr, BrokenAfterConnect())

    # Stop the server
    zc.ngi.wordcount.stop_server_process(zc.ngi.async.connect, addr)

    # Create a lister with a broken server and connect to it
    port = zc.ngi.wordcount.get_port()
    addr = 'localhost', port
    zc.ngi.async.listener(addr, BrokenConnect())
    zc.ngi.async.connect(addr, BrokenAfterConnect())

    # Create a lister with a broken Server handler and connect to it
    port = zc.ngi.wordcount.get_port()
    addr = 'localhost', port
    zc.ngi.async.listener(addr, BrokenAfterConnect())
    zc.ngi.async.connect(addr, BrokenAfterConnect())

def cleanup_async(test):
    zc.ngi.async.cleanup_map()
    zc.ngi.async.wait(1)

def test_suite():
    return unittest.TestSuite([
        manuel.testing.TestSuite(
            manuel.capture.Manuel() + manuel.doctest.Manuel(),
            'doc/index.txt',
            ),
        doctest.DocFileSuite(
            'old.test',
            'testing.test',
            'message.test',
            'adapters.test',
            'blocking.test',
            'async-udp.test',
            tearDown=cleanup_async),
        doctest.DocFileSuite(
            'async.test',
            setUp=async_evil_setup, tearDown=cleanup_async,
            ),
        doctest.DocTestSuite(setUp=cleanup_async, tearDown=cleanup_async),
        ])

if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')
